__author__ = "AlbertUnruh"
__url__ = "https://github.com/AlbertUnruh/AlbertUnruhUtils.py"
__license__ = "MIT"
__copyright__ = f"(c) 2021 - present {__author__}"
__version__ = "2.2.0"
__description__ = "A collection of utils written in Python."

from . import config
from . import ratelimit
from . import utils
from . import visual

from .config import *
from .ratelimit import *
from .utils import *
from .visual import *
